/***************************************************************************
* PostTrain.C															   *
*																		   *
* Functions for generating "Feature Matrix" files from raw training data.  *
* This must be performed after new training files have been added or	   *
* existing ones have been modified, since the feature recognition code	   *
* reads only the feature matrices, not the training data.				   *
*																		   *
*	Changes:															   *
*																		   *
*		08/15/99	 arvo	 Removed all fixed-length arrays.			   *
*		08/06/99	 arvo	 Major cleanup. 							   *
*		01/05/97	 guo	 Initial codiing.							   *
*																		   *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, James Arvo										   *
*																		   *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html	   *
*																		   *
* This program is distributed in the hope that it will be useful, but	   *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more	   *
* details.																   *
*																		   *
***************************************************************************/
//#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <tcl.h>
#include <tk.h>
#include <windows.h>
#include <fstream>
using namespace std;
#include "../util/Char.h"
#include "../util/List.h"
#include "../util/Args.h"
#include "../util/DirectoryTree.h"
#include "../rec/Font.h"
#include "../rec/PostTrain.h"
//#include <stdio.h>
//#include "../rec/Train.h"

extern ofstream cout, cerr;

static int 
total_symbols  = 0,
total_matrices = 0;
const char *data_dir = "data\\";
const char *font_dir = "fonts\\";
const char *data_file = "data.cit";

static void tab( int level )
{
	for( int i = 0; i < level; i++ ) cout << "  ";
}

/***************************************************************************
* Given a list of FeatureStatVectors, this function writes the information
* to a file called "FeatureMatrix" in the given directory.	The FeatureMatrix
* is intended to encode all the information embodied in the training files
* found within the same directory.	Thus, the feature-based recognizer need
* only read the FeatureMatrix files, and not the raw training data.
***************************************************************************/
void WriteFeatureMatrix( const char *dir_name, const List &feature_stats )
{
	filebuf fbuff;
	char filename[ Max_Name_Length ];
	int  entries = feature_stats.NumEntries();

	sprintf( filename, "%s/FeatureMatrix", dir_name );
	fbuff.open( filename, ios::out );
	ostream out( &fbuff );

	out << "[Feature Matrix: Version " << FM_Version << "]" << endl << entries << endl << endl;

	for( ListItem *item = feature_stats.First(); item != NULL; item = item->Next() ) 
	{
		FeatureStatsVector *fsv = (FeatureStatsVector *)item;
		out << *fsv;
		out << endl;
	}

	cout << "Encoded " << entries << " symbols." << endl << endl;
	total_symbols  += entries;
	total_matrices += 1;
	fbuff.close();
}

/***************************************************************************
* Takes all the feature matrices from all the subdirectories of the given
* directory and combines them into a single feature matrix in the root
* directory with file name data_file
***************************************************************************/
void CompactFeatureMatrices( const char *dir_name, const char *file_name )
{
	char tmp[1000];
	ifstream infile;
	ofstream outfile;
	DirectoryTree tree;
	DT_File *file;

	tree.ScanDirectory(dir_name);
	file = tree.GetFirstFile();
	cout << "Writing " << file_name << endl;
	outfile.open(file_name);
	outfile << "[Feature Matrix: Version " << FM_Version << "]" << endl << total_symbols << endl;

	while (file != NULL)
	{
		if (strcmp(file->name, "FeatureMatrix") == 0)
		{
			infile.open(file->GetFullName());
			infile.getline(tmp, 1000);
			infile.getline(tmp, 1000);
			while (!infile.eof())
			{
				infile.getline(tmp, 1000);
				outfile << tmp << endl;
			}
			infile.close();
		}
		file = tree.GetNextFile();
	}
	outfile.close();
}

/***************************************************************************
* Read in all training files in each directory and compute a "FeatureMatrix"
* file that captures the characterizes the statistical variation of each
* feature within the samples of each symbol (and style); these feature 
* matrices are in turn used by the recognizer, rather than the original 
* training data.  A FeatureMatrix represents all the training files found
* in a SINGLE directory; thus, many feature matrices may be generated by
* this function -- one for each sub-directory.
***************************************************************************/
static void ComputeFeatureMatrix( const char *dir_name )
{
	char *buff, *fname;
	total_symbols  = 0;
	total_matrices = 0;
	DirectoryTree tree;
	tree.ScanDirectory(dir_name);
	DT_File *file = tree.GetFirstFile();
	List feature_stats;

	cout << "\n-------------------------------------------------------------\n";  
	cout << "Beginning feature extraction at "<< dir_name << endl << endl;

	// Process all files in this directory.  No recursion!
	while (file != NULL)
	{
		buff = file->dir->GetFullName();
		cout << "Directory " << buff << ":" << endl;

		feature_stats.Clear();
		do
		{
			if (strcmp(file->name, "FeatureMatrix") == 0)
			{
				file = tree.GetNextFile();
				continue;
			}
			cout << "file " << file->name << ": ";
			FeatureStatsVector *fsv = new FeatureStatsVector;
			fname = file->GetFullName();
			tree.ConvertBackslashes(fname);
			fsv->ProcessTrainingFile(fname);
			feature_stats << fsv;
			file = tree.GetNextFile();
		} while (!tree.ChangedDir());

		if( feature_stats.NumEntries() > 0 ) 
		{
			WriteFeatureMatrix( buff, feature_stats );
		}
	}

	cout << "End of feature extraction at " << dir_name << endl;
	cout << "Encoded a total of " << total_symbols << " symbols in " 
		<< total_matrices << " FeatureMatrix file(s)." << endl;
	cout << "-------------------------------------------------------------";
	cout << endl << endl;
}

/***************************************************************************
* Main program for PostTrain.  Handle command-line arguments, then
* start recursive computation of feature matrices in all sub-directories.
****************************************************************************/
int main( int argc, char *argv[] )
{
	// Make the command line arguments available to any module
	// that wants to interrogate them.
	CommandLine args( argc, argv );
	char tmp[300];
	char *dir_name = new char[256];
	strcpy(dir_name, data_dir);

	cout.open("cout.txt");
	cerr.open("cerr.txt");

	// Handle the "version" flag.
	if( args.GetSwitch( "v" ) )
	{
		cout << "Version of Software       = " << SoftwareVersion << endl;
		cout << "Version of Training Files = " << TD_Version	  << endl;
		cout << "Version of Feature Matrix = " << FM_Version	  << endl;
		exit(0);
	}
	
	// Handle the "file" flag.
	if( args.GetSwitch( "d" ) )
	{
		delete dir_name;
		dir_name = (char *)args.GetArg( "d" );
	}
	
	// Handle the "help" flag.
	if( args.GetSwitch( "h" ) )
	{
		cout << endl;
		cout << "Usage: PostTrain { -[dhv] [option] }*" << endl 		<< endl;
		cout << "  -d <dname>  Read training data from this directory." << endl;
		cout << "  -h          Print these instructions and quit."		<< endl;
		cout << "  -v          Print the version number and quit."		<< endl;
		cout << endl;
		exit(0);
	}

	strcpy(tmp, dir_name);
	strcat(tmp, data_file);
	DeleteFile(tmp);
	ComputeFeatureMatrix( dir_name );
	CompactFeatureMatrices( dir_name, data_file );
	return 0;
}

#include "stdafx.h"
