/***************************************************************************
* TCL.C 																   *
*																		   *
* The TCL class and associated macros provide an easy method for creating  *
* a tcl/tk application that communicates with C++ code.  The macro called  *
* "DefineTclCallback" defines a C++ function that can be invoked from	   *
* tcl/tk via the "SendMessage" command.  That is, all communication is	   *
* funneled through one tcl command, which handles all C++ callbacks.	   *
*																		   *
*	Changes:															   *
*																		   *
*		01/09/2000	arvo  Initial coding.								   *
*																		   *
*--------------------------------------------------------------------------*
* Copyright (C) 2000, James Arvo										   *
*																		   *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html	   *
*																		   *
* This program is distributed in the hope that it will be useful, but	   *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more	   *
* details.																   *
*																		   *
***************************************************************************/
#include <assert.h>
#include <string.h>
#include "../util/TCL.h"
#include "../util/TCL_Thread.h"

extern ofstream cerr, cout;

TCL 		TCL::Cmd;
const char *TCL::End = "end";


int cmd_len = 0;

static const int
cmd_buff_size = 2048,
buffer_size   = 2048;

static char 
cmd_buff[ cmd_buff_size ],
buffer[ buffer_size ],
tags_buff[128],
fill_buff[128], 
line_buff[128], 
width_buff[128],
curv_buff[128];

class list_item {
public:
	list_item( const char *name_, Tcl_Callback func_ );
	const char *Name() const { return name; }
	list_item  *Next() const { return next; }
	Tcl_Callback Func() const { return func; }
	int Func( int argc,const char *argv[] ) const;
	static list_item *list_item::Lookup( const char  *name );
	static list_item *list_item::Lookup( Tcl_Callback func );
private:
	static list_item *root;
	const char	 *name;
	Tcl_Callback  func;
	list_item	 *next;
	int 		  wait;
};


list_item::list_item( const char *name_, Tcl_Callback func_ )
{
	name = strdup( name_ );
	func = func_;
	next = root;
	root = this;
	wait = 0;
}

int list_item::Func( int argc, const char *argv[] ) const 
{
	return func( argc, argv ); 
}

list_item *list_item::Lookup( const char *name )
{
	for( list_item *item = root; item != NULL; item = item->Next() )
	{
		if( strcmp( name, item->Name() ) == 0 )
		{
			return item;
		}
	}
	return NULL;
}

list_item *list_item::Lookup( Tcl_Callback func )
{
	for( list_item *item = root; item != NULL; item = item->Next() )
	{
		if( func == item->Func() )
		{
			return item;
		}
	}
	return NULL;
}

char		*TCL::tcl_script = NULL;
Tcl_Interp	*TCL::interp	 = NULL;
Tcl_UserInit TCL::user_init  = NULL;
int 		 TCL::width 	 = 1;
char		 TCL::canv[64]	 = "";
char		 TCL::fill[64]	 = "black";
char		 TCL::line[64]	 = "black";
list_item	*list_item::root = NULL;

/***************************************************************************
* Check that the version of Tcl/Tk is recent enough, and initialize all
* data members.  In particular, set the "base directory" to either the
* supplied path, or to "./" -- this is where to look for the Tcl/Tk file(s).
* The "data directory" is where to look for training files and/or feature
* matrices.
**************************************************************************/
TCL::TCL()
{
}

TCL::~TCL()
{
}

/***************************************************************************
* Start up a Tcl interpreter and pass it the name of the C++ function
* that will initialize Tcl & Tk, create new Tcl command, and finally
* start the interface running.	CONTROL NEVER RETURNS FROM "RUN".
**************************************************************************/
void TCL::Run( const char *path_to_script, const char *banner, Tcl_UserInit init )
{
	user_init = init;
	tcl_script = NULL;
	if( TCL_MAJOR_VERSION < 8 )
	{
		cerr << "Tcl/Tk version 8.0 or higher is required by this application." << endl;
		cerr << "The installed version is " << TCL_MAJOR_VERSION << endl;
		exit(0);
	}
	
	tcl_script = strdup( path_to_script );
	
	static char *argv[1];
	if( banner == NULL ) banner = "Tcl application";
	argv[0] = strdup( banner );
	Tk_Main( 1, argv, TCL::Startup );
	// Control never returns from Tk_Main.
	assert(0);
}

/***************************************************************************
* This function is called by Tk_Main after creating an interpreter but
* before starting the event loop.  This is where any special initialization
* of the Tcl environment must be performed -- such as establishing the
* "SendMessage" command that Tcl will use to communicate with C++.
**************************************************************************/
int TCL::Startup( Tcl_Interp *interp_ptr )
{
	interp = interp_ptr;
	
	if( Tcl_Init( interp ) != TCL_OK ) 
	{
		cerr << "Warning! Tcl_Init failed!\n";
		exit(0);
	}
	
	if( Tk_Init( interp ) != TCL_OK ) 
	{
		cerr << "Warning! Tk_Init failed!\n";
		exit(0);
	}
	
	// Create only one new command for the Tcl environment.  All further
	// dispatching is done at this end.
	
	Tcl_CreateCommand( interp, "SendMessage", SendMessage, NULL, NULL );
	
	// Tell Tcl where to find the ".tcl" script to run.  This was specified
	// in the constructor for this class.
	
	Tcl_SetVar( interp, "tcl_rcFileName" , tcl_script, 0 );
	
	// Initialize the Thread class with the newly-created Tcl interpreter,
	// then call the usser-supplied startup procedure (if there is one).
	
	TCL_Thread::SetInterp( interp );
	if( user_init != NULL ) user_init( interp );
	
	return TCL_OK;
}

int TCL::AddCallback( const char *name, Tcl_Callback func )
{
	list_item *item = new list_item( name, func );
	// return a dummy value so that "AddCallback" can be used to initialize
	// static variables, which will ensure that it is called BEFORE main is
	// invoked.  This is a trick to get the table of callbacks set up before
	// Tcl/Tk is even started up.
	return 1;
}

void TCL::Command( const char *cmd1, const char *cmd2, const char *cmd3 )
{
	Tcl_VarEval( interp, cmd1, " ", cmd2, " ", cmd3, NULL );
}

void TCL::Command( const char *cmd, int arg )
{
	sprintf( buffer, "%s %d", cmd, arg );
	Tcl_VarEval( interp, buffer, NULL );
}

void TCL::Command( const char *cmd, int arg1, int arg2 )
{
	sprintf( buffer, "%s %d %d", cmd, arg1, arg2 );
	Tcl_VarEval( interp, buffer, NULL );
}

TCL &TCL::operator<<( const char *cmd )
{
	if( cmd == TCL::End )
	{
		Tcl_VarEval( interp, cmd_buff, NULL );
		cmd_len = 0;
		return *this;
	}
	else if( strlen( cmd ) + cmd_len >= cmd_buff_size )
	{
		Tcl_VarEval( interp, cmd_buff, NULL );
		cmd_len = 0;
	}
	cmd_len += sprintf( cmd_buff + cmd_len, "%s ", cmd );
	return *this;
}

TCL &TCL::operator<<( int value )
{
	if( cmd_len >= cmd_buff_size - 10 )
	{
		Tcl_VarEval( interp, cmd_buff, NULL );
		cmd_len = 0;
	}
	cmd_len += sprintf( cmd_buff + cmd_len, "%d ", value );
	return *this;
}

int TCL::BadArgs( int argc, const char ** )
{
	const char *name = "<Unknown>";
	sprintf( interp->result, "%s: Wrong number of arguments (%d)", name, argc );
	return TCL_ERROR;
}

int TCL::Return( const char *value )
{
	char buff[128];
	strcpy( buff, value );
	Tcl_SetResult( interp, buff, TCL_VOLATILE );
	return TCL_OK;
}

void TCL::Draw( const Vec2 &P, const Vec2 &Q, const char *tag )
{
	if( tag == NULL ) tags_buff[0] = '\0';
	else sprintf( tags_buff, "-tag %s", tag );
	
	if( GetWidth() == 1 ) width_buff[0] = '\0';
	else sprintf( width_buff, "-width %d", GetWidth() );
	
	if( strcmp( GetLine(), "black" ) == 0 ) line_buff[0] = '\0';
	else sprintf( line_buff, "-fill %s", GetLine() );
	
	sprintf( buffer, "%s create line %d %d %d %d %s %s %s", 
		GetCanvas(),
		(int)P.X(), (int)P.Y(), (int)Q.X(), (int)Q.Y(),
		tags_buff, line_buff, width_buff
		);
	Command( buffer );
}

void TCL::Draw( const Multiline &line, const char *tag, int smoothness, int arrow )
{
	char arrow_buff[64];
	
	if( tag == NULL ) tags_buff[0] = '\0';
	else sprintf( tags_buff, "-tag %s", tag );
	
	if( GetWidth() == 1 ) width_buff[0] = '\0';
	else sprintf( width_buff, "-width %d", GetWidth() );
	
	if( strcmp( GetLine(), "black" ) == 0 ) line_buff[0] = '\0';
	else sprintf( line_buff, "-fill %s", GetLine() );
	
	if( smoothness == 0 ) curv_buff[0] = '\0';
	else sprintf( curv_buff, "-smooth true -splinesteps %d", smoothness );
	
	if( arrow == 0 ) arrow_buff[0] = '\0';
	else sprintf( arrow_buff, "-arrow last -arrowshape {18 22 4}" );
	
	int len = sprintf( buffer, "%s create line ", GetCanvas() );
	int count = 0;
	Pnode *prev = NULL;
	
	for( Pnode *n = line.First(); n != NULL; n = n->Next() )
	{
		if( n->Jump() && n != line.First() )
		{
			sprintf( buffer + len, "%s %s %s %s %s", 
				tags_buff, fill_buff, line_buff, width_buff, curv_buff );
			Command( buffer );
			len = sprintf( buffer, "%s create line ", GetCanvas() );
			count = 0;
		}
		else if( count >= 100 ) // Send a portion of the multiline.
		{
			sprintf( buffer + len, "%s %s %s %s %s", 
				tags_buff, fill_buff, line_buff, width_buff, curv_buff );
			Command( buffer );
			len = sprintf( buffer, "%s create line %1.0f %1.0f ", GetCanvas(), prev->X(), prev->Y() );
			count = 1;
		}
		assert( len < sizeof(buffer) - 20 );
		len += sprintf( buffer + len, "%1.0f %1.0f ", n->X(), n->Y() );
		prev = n;
		count++;
	}
	if( count > 1 ) 
	{
		sprintf( buffer + len, "%s %s %s %s %s %s", 
			tags_buff, fill_buff, line_buff, width_buff, curv_buff, arrow_buff );
		Command( buffer );
	}
}

void TCL::Draw( const Circle &C, const char *tag )
{
	float x1 = C.Center().X() - C.Rad();
	float y1 = C.Center().Y() - C.Rad();
	float x2 = C.Center().X() + C.Rad();
	float y2 = C.Center().Y() + C.Rad();
	
	if( tag == NULL ) tags_buff[0] = '\0';
	else sprintf( tags_buff, "-tag %s", tag );
	
	if( GetWidth() == 1 ) width_buff[0] = '\0';
	else sprintf( width_buff, "-width %d", GetWidth() );
	
	if( strcmp( GetFill(), "black" ) == 0 ) fill_buff[0] = '\0';
	else sprintf( fill_buff, "-fill %s", GetFill() );
	
	if( strcmp( GetLine(), "black" ) == 0 ) line_buff[0] = '\0';
	else sprintf( line_buff, "-outline %s", GetLine() );
	
	sprintf( buffer, "%s create oval %d %d %d %d %s %s %s %s",
		GetCanvas(),
		(int)x1, (int)y1, 
		(int)x2, (int)y2,
		tags_buff,
		fill_buff,
		line_buff,
		width_buff
		);
	Command( buffer );
}

void TCL::Draw( const Box2d &B, const char *tag )
{
	Multiline box;
	box << B.NW() << B.NE() << B.SE() << B.SW() << B.NW();
	Draw( box, tag );
}

void TCL::Marker( const Vec2 &P, int rad, const char *color, const char *tag )
{
	float x1 = P.X() - rad;
	float y1 = P.Y() - rad;
	float x2 = P.X() + rad;
	float y2 = P.Y() + rad;
	
	if( tag == NULL ) tags_buff[0] = '\0';
	else sprintf( tags_buff, "-tag %s", tag );
	
	sprintf( buffer, "%s create oval %d %d %d %d %s -fill %s",
		GetCanvas(),
		(int)x1, (int)y1, 
		(int)x2, (int)y2,
		tags_buff,
		color
		);
	Command( buffer );
}

void TCL::Translate( const Vec2 &trans, const char *tag )
{
	if( tag == NULL ) return;
	
	sprintf( buffer, "%s move %s %d %d", 
		GetCanvas(),
		tag,
		(int)trans.X(), 
		(int)trans.Y()
		);
	Command( buffer );
}

void TCL::Lower( const char *tag )
{
	sprintf( buffer, "%s lower %s", GetCanvas(), tag );
	Command( buffer );
}

void TCL::Raise( const char *tag )
{
	sprintf( buffer, "%s raise %s", GetCanvas(), tag );
	Command( buffer );
}

void TCL::Delete( const char *tag )
{
	sprintf( buffer, "%s delete {%s}", GetCanvas(), tag );
	Command( buffer );
}

void TCL::AddTag( const char *new_tag, const char *old_tag )
{
	if( old_tag == NULL )  
		sprintf( buffer, "%s addtag %s all" 	  , GetCanvas(), new_tag		  );
	else sprintf( buffer, "%s addtag %s withtag %s", GetCanvas(), new_tag, old_tag );
	Command( buffer );
}

/***************************************************************************
* This is the only C++ procedure that is invoked from Tcl.	From here
* the appropriate procedures are called by dispaching on the first 
* argument, which is assumed to be the name of the C++ function to call.
* The remaining arguments (if there are any) are then passed to that
* C++ function via "argc" and "argv".
**************************************************************************/
int TCL::SendMessage( ClientData, Tcl_Interp *, int argc, const char *argv[] )
{
	const char *name = argv[1];
	if( argc < 2 ) 
	{
		cerr << "ERROR: no args passed to SendMessage." << endl;
		return TCL_ERROR;
	}
	// Look up the corresponding function pointer and invoke the function.
	list_item *item = list_item::Lookup( name );
	if( item == NULL )
	{
		// A function by this name has not been defined.
		sprintf( interp->result, "Undefined function: %s", name );
		return TCL_ERROR;
	}
	else
	{
		// When invoking the function, strip off argv[0], which is "SendMessage".
		// Then the new argv[0] will be the name of this function, according to
		// the usual convention.
		item->Func( argc - 1, argv + 1 );
	}
	return TCL_OK;
}

#include "stdafx.h"
