/***************************************************************************
* TCL.h                                                                    *
*                                                                          *
* The TCL class and associated macros provide an easy method for creating  *
* a tcl/tk application that communicates with C++ code.  The macro called  *
* "DefineTclCallback" defines a C++ function that can be invoked from      *
* tcl/tk via the "SendMessage" command.  That is, all communication is     *
* funneled through one tcl command, which handles all C++ callbacks.       *
*                                                                          *
*   Changes:                                                               *
*                                                                          *
*       09/03/2000  arvo  Added support for static methods as callbacks.   *
*       01/09/2000  arvo  Initial coding.                                  *
*                                                                          *
*--------------------------------------------------------------------------*
* Copyright (C) 2000, James Arvo                                           *
*                                                                          *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html      *
*                                                                          *
* This program is distributed in the hope that it will be useful, but      *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more     *
* details.                                                                 *
*                                                                          *
***************************************************************************/
#ifndef __TCL_INCLUDED__
#define __TCL_INCLUDED__

//#include <iostream.h>
#include <tcl.h>
#include <tk.h>
#include "../util/Vec2.h"
#include "../util/Box2.h"
#include "../util/Circle.h"
#include "../util/Multiline.h"

// TCL_CALLBACK is used to declare a function that will be called from TCL
// (without inserting the name into the callback table).  This will not 
// typically be needed in user-level code, except when the function is
// a static member of a class.
//
#define TCL_CALLBACK( Name ) int Name( int argc, char *argv[] )

// DefineTclCallback is used for defining a C++ function to be called from TCL.  
// It ensures that the function has the appropriate signature, and also has the
// side effect of entering the name into a table of functions that can be invoked
// from TCL using the "SendMessage" command.  (Note: the "dummy_init_var" variable
// is used to force the callback to be added before the main program is executed.)
//
#define DefineTclCallback( Name )         \
    extern TCL_CALLBACK( Name );          \
    static int dummy_init_var_##Name =    \
        TCL::AddCallback( #Name, Name );  \
    TCL_CALLBACK( Name )

#define DefineTclCallbackMethod( Class, Name )   \
    static int dummy_##Class##_method_##Name =   \
        TCL::AddCallback( #Name, Class::Name );  \
    TCL_CALLBACK( Class::Name )

typedef int  (*Tcl_Callback)( int argc, const char *argv[] );
typedef void (*Tcl_UserInit)( Tcl_Interp *interp );

// The TCL class provides a convenient encapsulation of all the code needed to
// make Tcl and C++ work together.  All of the methods are static so that pointers
// to some of them can be passed to Tcl, and so that all are easily accessible
// to C++ functions that wish to draw on a Tcl/Tk canvas (e.g. an instance of
// this class need not be made available to such functions).

class TCL {

  public: // Fundamental public methods.  Here, "Run" is the most important.
    TCL();
   ~TCL();
    static void Run( const char *script, const char *banner = NULL, Tcl_UserInit = NULL );
    static int  AddCallback( const char *name, Tcl_Callback func );      // Used by macro.
    static int  Running() { return interp != NULL; }                     // Has Tcl been started?

  public: // Basic utilities called from within the user-defined functions.
    static Tcl_Interp *GetInterp() { return interp; }
    static void Command( const char *, const char * = NULL, const char * = NULL );
    static void Command( const char *, int      );
    static void Command( const char *, int, int );
    static int  BadArgs( int argc, const char *argv[] ); // Reports an error.
    static int  Return ( const char * );

  public: // State-querying methods.
    static int         GetWidth  () { return width; }
    static const char *GetCanvas () { return canv;  }
    static const char *GetFill   () { return fill;  }
    static const char *GetLine   () { return line;  }

  public: // State-setting methods.
    static void SetWidth ( int         w ) { width = w;         }
    static void SetCanvas( const char *c ) { strcpy( canv, c ); }
    static void SetFill  ( const char *f ) { strcpy( fill, f ); }
    static void SetLine  ( const char *l ) { strcpy( line, l ); }

  public: // Drawing methods.
    static void Draw( const Vec2 &P, const Vec2 &Q, const char *tag = NULL );
    static void Draw( const Multiline &           , const char *tag = NULL, int = 0, int = 0 );
    static void Draw( const Circle    &           , const char *tag = NULL );
    static void Draw( const Box2d     &           , const char *tag = NULL );
    static void Marker( const Vec2 &, int rad, const char *color, const char *tag = NULL );

  public: // Editing methods.
    static void Translate( const Vec2 &, const char *tag  = NULL );
    static void Delete   ( const char *tag = NULL );
    static void Raise    ( const char *tag = NULL );
    static void Lower    ( const char *tag = NULL );
    static void AddTag   ( const char *new_tag, const char *old_tag = NULL );

  public: // Operators & constants for sending commands to Tcl.
    static TCL Cmd;
    static const char *End;
    TCL &operator<<( const char * ); // Alternative to "Command".
    TCL &operator<<( int );

  private: // Methods for initialization and for communicating with C++ from TCL.
    static int Startup( Tcl_Interp *interp );
    static int SendMessage( ClientData, Tcl_Interp *, int argc, const char *argv[] );

  private: // Private data.
    static Tcl_Interp *interp;      // The Tcl interpreter.
    static Tcl_UserInit user_init;  // Function to invoke after interpreter exists.
    static char *tcl_script;        // Where the Tcl/Tk script is.
    static int   width; 
    static char  canv[64];
    static char  fill[64];
    static char  line[64];
  };


#endif




