/***************************************************************************
* TCL_Thread.h                                                             *
*                                                                          *
* The TCL_Thread class is the base class for all objects that wish to be   *
* awakened periodically, via a "WakeUp" call.  Each TCL_Thread object can  *
* set its own wakeup period and timeslice.  Relinquishing control is on    *
* a purely voluntary basis (consequently, these are not true threads).     *
*                                                                          *
*   Changes:                                                               *
*                                                                          *
*       08/06/2000   arvo    Rewritten for TCL wrapper.                    *
*       04/20/1998   arvo    Initial coding.                               *
*                                                                          *
***************************************************************************/
#ifndef __TCL_THREAD_INCLUDED__
#define __TCL_THREAD_INCLUDED__

#include <tcl.h>


static const int
    UpdateTime   = 30,   // Milliseconds between updates.
    MinDelay     = 10,   // Minimum delay before callback. 
    MinTimeSlice = 20,   // Minimum time any thread is alotted.
    MaxTimeSlice = 100;  // Maximum time any thread is alotted.


// To create a thread, derive a class from TCL_Thread and supply
// appropriate virtual methods for "Start", "WakeUp", and "Stop".
// The constructor of the class should also establish whatever state
// is appropriate for the thread.  Typically, all Tcl calls will be
// made by the above methods using the static public methods of the 
// TCL class.
//
// To start a thread, call its StartThread method.  This will
// invoke its virtual Start method, followed by the virtual WakeUp
// method, and then continue to periodically call the WakeUp method,
// until the StopThread method is called.  When StopMethod is called,
// any pending wakeups are cancelled, and the virtual Stop method is
// called.  The virtual methods "Start" and "Stop" can be used for
// setup and cleanup, and need not be supplied at all (the base class
// is supplied with trivial versions).

class TCL_Thread {

    public:
        TCL_Thread();
        void   SetWakeupPeriod( int period );  // milliseconds
        void   SetTimeSlice   ( int slice  );  // milliseconds
        int    TimedOut() const;               // Time slice is used up.
        void   StartThread();                  // Turn on periodic wakeups.
        void   StopThread();                   // Turn off periodic wakeups.
        int    Running() const;                // Is the thread awake?
        int    Locked() const { return lock_count > 1; }

    public: // Virtual methods.
        virtual ~TCL_Thread();
        virtual void Start () {}   // Optionally redefine for descendants.
        virtual void WakeUp() = 0; // Must be define for descendants.
        virtual void Stop  () {}   // Optionally redefine for descendants.

    public: // Static methods.
        static long Now();
        static int  WakeUpThread( ClientData, Tcl_Interp *, int argc, const char **argv );
        static void SetInterp( Tcl_Interp * );

    private:
        void WakeUpWrapper(); // Called by WakeUpThread.
        void WakeUpAfter( int delay );
        int  running;       // Indicates whether thread is awake.
        int  lock_count;    // A semiphore for mutual exclusion.
        int  active;        // Wake up periodically if true.
        int  wakeup_period; // Milliseconds between wakeups.
        int  min_delay;     // Minimum wait before another wakeup.
        int  time_slice;    // Milliseconds allotted at wakeup.
        long alarm_time;    // Time at which alarm was reset.
        long start_time;    // Time at which thread is awakened.
        int  Lock()   { return ++lock_count; }
        int  Unlock() { return --lock_count; }
        static char cbid[64]; // ID of pending callback.
        static Tcl_Interp *interp;  // The Tcl interpreter.

    };


#endif



